@@ -1,8 +1,9 @@
 from curses import tigetnum, tigetstr, setupterm, tparm
 import inspect
+from itertools import cycle
 import logging
 import os
-from traceback import format_list, extract_tb
+from traceback import format_list, extract_tb, format_exception_only
 from unittest import TestCase
 
 from nose.plugins import Plugin
@@ -38,17 +39,38 @@ class ProgressivePlugin(Plugin):
             writeln = flush = write = lambda self, *args: None
 
         self.stream = stream
-        setupterm(None, stream.fileno())  # Make setupterm() work even when -s is passed. TODO: Don't do this if self.stream isn't a terminal. Use os.isatty(self.stream.fileno()). If it isn't, perhaps replace the ShyProgressBar with a dummy object.
-        self.bar = ProgressBar()
+        # Explicit args make setupterm() work even when -s is passed:
+        setupterm(None, stream.fileno())
+        # TODO: Don't call setupterm() if self.stream isn't a terminal. Use
+        # os.isatty(self.stream.fileno()). If it isn't, perhaps replace the
+        # ShyProgressBar with a dummy object.
+
         return DummyStream()
 
+    def prepareTest(self, test):
+        """Init the progress bar, and tell it how many tests there are."""
+        # Not really clear from the docs whether this always gets called in
+        # every plugin or whether another plugin can prevent this from getting
+        # called by returning something. If the latter, we'll be surprised when
+        # self.bar doesn't exist.
+        self.bar = ProgressBar(test.countTestCases())
+
     def printError(self, kind, err, test):
-        _, _, tb = err
-        extracted_tb = extract_tb(tb)
+        """Output a human-readable error report to the stream.
+
+        kind -- the (string) type of incident the precipitated this call
+        err -- exc_info()-style traceback triple
+        test -- the test that precipitated this call
+
+        """
+        # Don't bind third item to a local var; that can create circular refs
+        # which are expensive to collect. See the sys.exc_info() docs.
+        exception_type, exception_value = err[:2]
+        extracted_tb = extract_tb(err[2])
         formatted_err = ''.join(format_list(extracted_tb))
-        # TODO: Format the tb ourselves and eliminate the space-wasting
-        # "Traceback (most recent..." line to make up for the extra pathname
-        # line.
+        # TODO: Canonicalize the path to remove /kitsune/../kitsune nonsense.
+        # Don't relativize, though, as that hurts the ability to paste into
+        # running editors.
         with ShyProgressBar(self.stream, self.bar):
             writeln = self.stream.writeln
             write = self.stream.write
@@ -60,13 +82,20 @@ class ProgressivePlugin(Plugin):
             address = test.address()
             if address:
                 file_name = address[0]
-                line_num = extracted_tb[-1][1]
-                writeln(' ' * len(kind) +
-                        '  %s +%s' % (relative_source_path(file_name), line_num))
+                if file_name:
+                    line_num = extracted_tb[-1][1]
+                    writeln(' ' * len(kind) +
+                            '  %s +%s' % (relative_source_path(file_name), line_num))
 
             write(tigetstr('sgr0'))
+
+            # Traceback:
             self.stream.write(formatted_err)
 
+            # Exception:
+            self.stream.write(''.join(format_exception_only(exception_type, exception_value)))
+
+
     def printErrors(self):
         # The current summary doesn't begin with a \n.
         with ShyProgressBar(self.stream, self.bar):
@@ -74,8 +103,11 @@ class ProgressivePlugin(Plugin):
 
     def finalize(self, result):
         """Print counts of tests run."""
-        types = ['test', 'failure', 'error', 'skip']
-        values = [self.testsRun, len(result.failures), len(result.errors), len(result.skipped)]
+        types = ['test', 'failure', 'error']
+        values = [self.testsRun, len(result.failures), len(result.errors)]
+        if hasattr(result, 'skipped'):  # Absent if --no-skip is passed
+            types.append('skip')
+            values.append(len(result.skipped))
         msg = ', '.join('%s %s%s' % (v, t, 's' if v != 1 else '')
                         for t, v in zip(types, values))
 
@@ -113,20 +145,48 @@ class ProgressivePlugin(Plugin):
 
 
 class ProgressBar(object):
-    def __init__(self):
-        width = tigetnum('cols')
-        self.last = ' ' * width
+    SPINNER_CHARS = r'/-\|'
+
+    def __init__(self, max):
+        """max is the highest value I will attain. Must be >0."""
+        self.last = ''  # The contents of the previous progress line printed
+        self.max = max
+        self.spinner = cycle(self.SPINNER_CHARS)
 
     def get(self, test, number):
-        number = str(number)
+        """Return updated content for the progress bar.
+
+        At the moment, the graph takes a fixed width, and the test identifier
+        takes the rest of the row, truncated from the left to fit.
+
+        test -- the test being run
+        number -- how many tests have been run so far, including this one
+
+        """
+        # TODO: Play nicely with absurdly narrow terminals.
+
+        # Figure out graph:
+        # We cheat a bit and dedicate extra column to the spinner to make the
+        # logic simple, so the graph is just sliiiightly misproportional.
+        GRAPH_WIDTH = 14
+        num_markers = int(round(float(number) / self.max * GRAPH_WIDTH))
+        # If there are any markers, replace the last one with the spinner.
+        # Otherwise, have just a spinner:
+        markers = '=' * (num_markers - 1) + self.spinner.next()
+        graph = '[%s%s]' % (markers,
+                            ' ' * (GRAPH_WIDTH - len(markers)))
+
+        # Figure out the test identifier portion:
         test_path = python_path(test.test) + '.' + test.test._testMethodName
         width = tigetnum('cols')
-        cols_for_path = width - len(number) - 2  # 2 spaces between path and number
+        cols_for_path = width - len(graph) - 2  # 2 spaces between path & graph
         if len(test_path) > cols_for_path:
             test_path = test_path[len(test_path) - cols_for_path:]
         else:
             test_path += ' ' * (cols_for_path - len(test_path))
-        self.last = tigetstr('bold') + test_path + '  ' + number + tigetstr('sgr0')
+        
+        # Put them together, and let simmer:
+        self.last = tigetstr('bold') + test_path + '  ' + graph + tigetstr('sgr0')
         return self.last
 
 
@@ -156,7 +216,6 @@ class ShyProgressBar(object):
     def __enter__(self):
         """Erase the progress bar so bits of disembodied progress bar don't get scrolled up the terminal."""
         # My terminal has no status line, so we make one manually.
-        # Doing this each time gives us a hope of adjusting if people resize their terminals during test runs:
         with AtProgressBar(self.stream):
             self.stream.write(tigetstr('el'))  # erase to EOL
         self.stream.flush()
